# Copyright (c) 2022, Simon Magin (simakro)
# BSD-2 license
# see LICENSE file for license details

"""
NoTrAmp is a Tool for read-depth normalization and trimming of amplicon
 reads generated with long read technologies (ONT/PacBio).
"""

from .version import __version__


VERSION = __version__
